#!/usr/bin/env python
# encoding: utf-8


import torch
import torch.nn as nn


def weights_init(m):
    classname = m.__class__.__name__
    if classname.find('Conv') != -1:
        m.weight.data.normal_(0.0, 0.02)
    elif classname.find('BatchNorm') != -1:
        m.weight.data.normal_(1.0, 0.02)
        m.bias.data.fill_(0)


class DRGAN_encoder(nn.Module):
  """
  Class defining the decoder in the DR-GAN architecture

  **Parameters**

  image_size: tuple
    The dimension of the image (CxHxW)
  
  latent_dim: int
    The dimension of the encoded ID
  """
  def __init__(self, image_size, latent_dim):
    
    # conv2d(in_channels, out_channels (i.e. number of feature maps), kernel size, stride, padding)

    super(DRGAN_encoder, self).__init__()
    
    self.ngpu = 1
    n_fm_first = 64 # number of feature maps at the first layer
       
    self.main = nn.Sequential(
      # input is (nc) x 64 x 64, output is n_mfirst x 64/2 x 64/2
      nn.Conv2d(image_size[0], n_fm_first, 4, 2, 1, bias=False),
      nn.LeakyReLU(0.2, inplace=True),
      # state size. (n_fm_first) x 32 x 32
      nn.Conv2d(n_fm_first, n_fm_first * 2, 4, 2, 1, bias=False),
      nn.BatchNorm2d(n_fm_first * 2),
      nn.LeakyReLU(0.2, inplace=True),
      # state size. (n_fm_first*2) x 16 x 16
      nn.Conv2d(n_fm_first * 2, n_fm_first * 4, 4, 2, 1, bias=False),
      nn.BatchNorm2d(n_fm_first * 4),
      nn.LeakyReLU(0.2, inplace=True),
      # state size. (n_fm_first*4) x 8 x 8
      nn.Conv2d(n_fm_first * 4, n_fm_first * 8, 4, 2, 1, bias=False),
      nn.BatchNorm2d(n_fm_first * 8),
      nn.LeakyReLU(0.2, inplace=True),
      # state size. (n_fm_first*8) x 4 x 4 
      nn.Conv2d(n_fm_first * 8, (latent_dim), 4, 2, 1, bias=False),
      # average pooling
      nn.AvgPool2d(2, stride=1)
    )


  def forward(self, x):
    """
    Forward function for the encoder.

    **Parameters**

    x: pyTorch Variable
      The minibatch of images to encode.
    """    
    if isinstance(x.data, torch.cuda.FloatTensor) and self.ngpu > 1:
      output = nn.parallel.data_parallel(self.main, x, range(self.ngpu))
    else:
      output = self.main(x)

    return output


class DRGAN_decoder(nn.Module):
  """
  Class defining the decoder in the DR-GAN architecture

  **Parameters**

  image_size: tuple
    The dimension of the image (CxHxW)
  
  noise_dim: int
    The dimension of the noise
  
  latent_dim: int
    The dimension of the encoded ID
  
  conditional_dim: int
    The dimension of the conditioning variable
  """
  def __init__(self, image_size, noise_dim, latent_dim, conditional_dim):

    super(DRGAN_decoder, self).__init__()
       
    self.ngpu = 1 # usually, we don't have more than one GPU 
    n_fm_first = 64 # number of feature maps at the first layer

    self.main = nn.Sequential(
      # input is Z+ID+C , going into a convolution
      nn.ConvTranspose2d((noise_dim + latent_dim + conditional_dim), n_fm_first * 8, 4, 1, 0, bias=False),
      nn.BatchNorm2d(n_fm_first * 8),
      nn.ReLU(True),
      # state size. (n_fm_first*8) x 4 x 4
      nn.ConvTranspose2d(n_fm_first * 8, n_fm_first * 4, 4, 2, 1, bias=False),
      nn.BatchNorm2d(n_fm_first * 4),
      nn.ReLU(True),
      # state size. (n_fm_first*4) x 8 x 8
      nn.ConvTranspose2d(n_fm_first * 4, n_fm_first * 2, 4, 2, 1, bias=False),
      nn.BatchNorm2d(n_fm_first * 2),
      nn.ReLU(True),
      # state size. (n_fm_first*2) x 16 x 16
      nn.ConvTranspose2d(n_fm_first * 2, n_fm_first, 4, 2, 1, bias=False),
      nn.BatchNorm2d(n_fm_first),
      nn.ReLU(True),
      # state size. (n_fm_first) x 32 x 32
      nn.ConvTranspose2d(n_fm_first, image_size[0], 4, 2, 1, bias=False),
      nn.Tanh()
      # state size. (nc) x 64 x 64
    )

  def forward(self, z, y, f):
    """
    Forward function for the decoder.

    **Parameters**

    z: pyTorch Variable
      The minibatch of noise.

    y: pyTorch Variable
      The conditional one hot encoded vector for the minibatch.
    
    f: pyTorch Variable
      The encoded ID for the minibatch 
    """    
    decoder_input = torch.cat((z, y, f), 1)
    if isinstance(decoder_input.data, torch.cuda.FloatTensor) and self.ngpu > 1:
      output = nn.parallel.data_parallel(self.main, decoder_input, range(self.ngpu))
    else:
      output = self.main(decoder_input)
    
    return output

class DRGAN_discriminator(nn.Module):
  """
  Class defining the discriminator in the DR-GAN architecture

  **Parameters**

  image_size: tuple
    The dimension of the image (CxHxW)
  
  number_of_ids: int
    The number of identities in the DB
  
  conditional_dim: int
    The dimension of the conditioning variable
  """
  def __init__(self, image_size, number_of_ids, conditional_dim):
    
    super(DRGAN_discriminator, self).__init__()
   
    self.number_of_ids = number_of_ids
    self.ngpu = 1
    n_fm_first = 64
       
    self.main = nn.Sequential(
      # input is (nc) x 64 x 64
      nn.Conv2d(image_size[0], n_fm_first, 4, 2, 1, bias=False),
      nn.LeakyReLU(0.2, inplace=True),
      # state size. (n_fm_first) x 32 x 32
      nn.Conv2d(n_fm_first, n_fm_first * 2, 4, 2, 1, bias=False),
      nn.BatchNorm2d(n_fm_first * 2),
      nn.LeakyReLU(0.2, inplace=True),
      # state size. (n_fm_first*2) x 16 x 16
      nn.Conv2d(n_fm_first * 2, n_fm_first * 4, 4, 2, 1, bias=False),
      nn.BatchNorm2d(n_fm_first * 4),
      nn.LeakyReLU(0.2, inplace=True),
      # state size. (n_fm_first*4) x 8 x 8
      nn.Conv2d(n_fm_first * 4, n_fm_first * 8, 4, 2, 1, bias=False),
      nn.BatchNorm2d(n_fm_first * 8),
      nn.LeakyReLU(0.2, inplace=True),
      # state size. (n_fm_first*8) x 4 x 4
      nn.Conv2d(n_fm_first * 8, (number_of_ids + conditional_dim + 1), 4, 1, 0, bias=False),
      nn.Sigmoid()
    )

  def forward(self, x):
    """
    Forward function for the encoder.

    **Parameters**

    x: pyTorch Variable
      The minibatch of images to process.
    """    
    if isinstance(x.data, torch.cuda.FloatTensor) and self.ngpu > 1:
      output = nn.parallel.data_parallel(self.main, x, range(self.ngpu))
    else:
      output = self.main(x)

    return output.squeeze()
