#!/usr/bin/env python
# encoding: utf-8


""" Train a Conditional GAN 

Usage:
  %(prog)s [--noise-dim=<int>] [--conditional-dim=<int>] 
           [--batch-size=<int>] [--epochs=<int>] [--sample=<int>]
           [--output-dir=<path>] [--use-gpu] [--seed=<int>] [--verbose ...]

Options:
  -h, --help                    Show this screen.
  -V, --version                 Show version.
  -n, --noise-dim=<int>         The dimension of the noise [default: 100]
  -c, --conditional-dim=<int>   The dimension of the conditional variable [default: 13]
  -b, --batch-size=<int>        The size of your mini-batch [default: 64]
  -e, --epochs=<int>            The number of training epochs [default: 100] 
  -s, --sample=<int>            Save generated images at every 'sample' batch iteration [default: 100000000000] 
  -o, --output-dir=<path>       Dir to save the logs, models and images [default: ./cgan-multipie/] 
  -g, --use-gpu                 Use the GPU 
  -S, --seed=<int>              The random seed [default: 3] 
  -v, --verbose                 Increase the verbosity (may appear multiple times).

Example:

  To run the training process 

    $ %(prog)s --batch-size 64 --epochs 25 --output-dir drgan 

See '%(prog)s --help' for more information.

"""

import os, sys
import pkg_resources

import bob.core
logger = bob.core.log.setup("bob.learn.pytorch")

from docopt import docopt

version = pkg_resources.require('bob.learn.pytorch')[0].version

import numpy
import bob.io.base

# torch
import torch
import torch.nn as nn
import torch.optim as optim
import torchvision.transforms as transforms
import torchvision.utils as vutils
from torch.autograd import Variable

# data and architecture from the package
from bob.learn.pytorch.datasets import MultiPIEDataset
from bob.learn.pytorch.datasets import RollChannels
from bob.learn.pytorch.datasets import ToTensor
from bob.learn.pytorch.datasets import Normalize

from bob.learn.pytorch.architectures import weights_init

from bob.learn.pytorch.architectures import ConditionalGAN_generator as cgenerator
from bob.learn.pytorch.architectures import ConditionalGAN_discriminator as cdiscriminator
from bob.learn.pytorch.trainers import ConditionalGANTrainer as ctrainer

def main(user_input=None):
  
  # Parse the command-line arguments
  if user_input is not None:
      arguments = user_input
  else:
      arguments = sys.argv[1:]

  prog = os.path.basename(sys.argv[0])
  completions = dict(prog=prog, version=version,)
  args = docopt(__doc__ % completions,argv=arguments,version='Train conditional GAN (%s)' % version,)

  # verbosity
  verbosity_level = args['--verbose']
  bob.core.log.set_verbosity_level(logger, verbosity_level)

  # get the arguments
  noise_dim = int(args['--noise-dim'])
  conditional_dim = int(args['--conditional-dim'])
  batch_size = int(args['--batch-size'])
  epochs = int(args['--epochs'])
  sample = int(args['--sample'])
  output_dir = str(args['--output-dir'])
  seed = int(args['--seed'])
  use_gpu = bool(args['--use-gpu'])

  images_dir = os.path.join(output_dir, 'samples')
  log_dir = os.path.join(output_dir, 'logs')
  model_dir = os.path.join(output_dir, 'models')

  # process on the arguments / options
  torch.manual_seed(seed)
  if use_gpu:
    torch.cuda.manual_seed_all(seed)
  if torch.cuda.is_available() and not use_gpu:
    logger.warn("You have a CUDA device, so you should probably run with --use-gpu")
  bob.io.base.create_directories_safe(images_dir)
  bob.io.base.create_directories_safe(log_dir)
  bob.io.base.create_directories_safe(images_dir)

  # ============
  # === DATA ===
  # ============
  
  # WARNING with the transforms ... act on labels too, at some point, I may have to write my own
  # Also, in 'ToTensor', there is a reshape performed from: HxWxC to CxHxW
  face_dataset = MultiPIEDataset(#root_dir='/Users/guillaumeheusch/work/idiap/data/multipie-cropped-64x64', 
                                 root_dir='/idiap/temp/heusch/data/multipie-cropped-64x64', 
                                 frontal_only=False, 
                                 transform=transforms.Compose([
                                   RollChannels(), # bob to skimage:  
                                   ToTensor(),
                                   Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
                                   ])
                                 )

  dataloader = torch.utils.data.DataLoader(face_dataset, batch_size=batch_size, shuffle=True)
  logger.info("There are {} training images".format(len(face_dataset))) 
  
  # ===============
  # === NETWORK ===
  # ===============
  ngpu = 1 # usually we don't have more than one GPU
  
  generator = cgenerator(noise_dim, conditional_dim)
  generator.apply(weights_init)
  logger.info("Generator architecture: {}".format(generator))

  discriminator = cdiscriminator(conditional_dim)
  discriminator.apply(weights_init)
  logger.info("Discriminator architecture: {}".format(discriminator))

  # ===============
  # === TRAINER ===
  # ===============
  trainer = ctrainer(generator, discriminator, [3, 64, 64], batch_size=batch_size, noise_dim=noise_dim, conditional_dim=conditional_dim, use_gpu=use_gpu, verbosity_level=verbosity_level)
  trainer.train(dataloader, n_epochs=epochs, output_dir=output_dir)
