#!/usr/bin/env python
# encoding: utf-8


""" Train a DR-GAN 

Usage:
  %(prog)s [--noise-dim=<int>] 
           [--batch-size=<int>] [--epochs=<int>] [--sample=<int>]
           [--output-dir=<path>] [--use-gpu] [--seed=<int>] [--verbose ...]

Options:
  -h, --help                Show this screen.
  -V, --version             Show version.
  -n, --noise-dim=<int>     the dimension of the noise [default: 100]
  -b, --batch-size=<int>    The size of your mini-batch [default: 64]
  -e, --epochs=<int>        The number of training epochs [default: 100] 
  -s, --sample=<int>        Save generated images at every 'sample' batch iteration [default: 100000000000] 
  -o, --output-dir=<path>   Dir to save the logs, models and images [default: ./dcgan-multipie/] 
  -g, --use-gpu             Use the GPU 
  -S, --seed=<int>          The random seed [default: 3] 
  -v, --verbose             Increase the verbosity (may appear multiple times).

Example:

  To run the training process 

    $ %(prog)s --batch-size 64 --epochs 25 --output-dir dcgan 

See '%(prog)s --help' for more information.

"""

import os, sys
import pkg_resources

import bob.core
logger = bob.core.log.setup("bob.learn.pytorch")

from docopt import docopt

version = pkg_resources.require('bob.learn.pytorch')[0].version

import numpy
import bob.io.base

# torch
import torch
import torch.nn as nn
import torch.optim as optim
import torchvision.transforms as transforms
import torchvision.utils as vutils
from torch.autograd import Variable

# data, architecture and trainer from the package
from bob.learn.pytorch.datasets import MultiPIEDataset
from bob.learn.pytorch.datasets import RollChannels
from bob.learn.pytorch.datasets import ToTensor
from bob.learn.pytorch.datasets import Normalize

from bob.learn.pytorch.architectures import DCGAN_generator
from bob.learn.pytorch.architectures import DCGAN_discriminator
from bob.learn.pytorch.architectures import weights_init

from bob.learn.pytorch.trainers import DCGANTrainer


def main(user_input=None):
  
  # Parse the command-line arguments
  if user_input is not None:
      arguments = user_input
  else:
      arguments = sys.argv[1:]

  prog = os.path.basename(sys.argv[0])
  completions = dict(prog=prog, version=version,)
  args = docopt(__doc__ % completions,argv=arguments,version='Train DCGAN (%s)' % version,)

  # verbosity
  verbosity_level = args['--verbose']
  bob.core.log.set_verbosity_level(logger, verbosity_level)

  # get the arguments
  noise_dim = int(args['--noise-dim'])
  batch_size = int(args['--batch-size'])
  epochs = int(args['--epochs'])
  sample = int(args['--sample'])
  output_dir = str(args['--output-dir'])
  seed = int(args['--seed'])
  use_gpu = bool(args['--use-gpu'])

  # process on the arguments / options
  torch.manual_seed(seed)
  if use_gpu:
    torch.cuda.manual_seed_all(seed)
  if torch.cuda.is_available() and not use_gpu:
    logger.warn("You have a CUDA device, so you should probably run with --use-gpu")
  
  bob.io.base.create_directories_safe(output_dir)

  # ============
  # === DATA ===
  # ============
  face_dataset = MultiPIEDataset(root_dir='/idiap/temp/heusch/data/multipie-cropped-64x64', 
                                 frontal_only=True, 
                                 transform=transforms.Compose([
                                   RollChannels(), # bob to skimage:  
                                   ToTensor(),
                                   Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
                                   ])
                                 )
  dataloader = torch.utils.data.DataLoader(face_dataset, batch_size=batch_size, shuffle=True)
  logger.info("There are {} training images".format(len(face_dataset))) 
  
  # ===============
  # === NETWORK ===
  # ===============
  ngpu = 1 # usually we don't have more than one GPU
  
  generator = DCGAN_generator(ngpu)
  generator.apply(weights_init)
  logger.info("Generator architecture: {}".format(generator))

  discriminator = DCGAN_discriminator(ngpu)
  discriminator.apply(weights_init)
  logger.info("Discriminator architecture: {}".format(discriminator))


  # ===============
  # === TRAINER ===
  # ===============
  trainer = DCGANTrainer(generator, discriminator, batch_size=batch_size, noise_dim=noise_dim, use_gpu=use_gpu, verbosity_level=verbosity_level)
  trainer.train(dataloader, n_epochs=epochs, output_dir=output_dir)
