#!/usr/bin/env python
# encoding: utf-8

import numpy
import time

import torch
import torch.nn as nn
import torch.optim as optim
from torch.autograd import Variable
import torchvision.utils as vutils

import bob.core
logger = bob.core.log.setup("bob.learn.pytorch")

import bob.io.base
import bob.io.image

class DRGANTrainer(object):
  """
  Class to train a  DR-GAN

  **Parameters**

  encoder: pytorch nn.Module
    The encoder network

  decoder: pytorch nn.Module
    The decoder network

  discriminator: pytorch nn.Module
    The discriminator network

  image_size: list
    The size of the images in this format: [channels,height, width]

  batch_size: int
    The size of your minibatch

  noise_dim: int
    The dimension of the noise (input to the generator)

  conditional_dim: int
    The dimension of the conditioning variable

  latent_dim: int
    The dimension of the encoded ID

  use_gpu: boolean
    If you would like to use the gpu

  verbosity_level: int
    The level of verbosity output to stdout
  """
  def __init__(self, encoder, decoder, discriminator, image_size, batch_size=64, noise_dim=100, conditional_dim=13, latent_dim=320, use_gpu=False, verbosity_level=2):
   
    bob.core.log.set_verbosity_level(logger, verbosity_level)
    
    self.encoder = encoder
    self.decoder = decoder
    self.discriminator = discriminator
    self.image_size = image_size
    self.batch_size = batch_size
    self.noise_dim = noise_dim 
    self.conditional_dim = conditional_dim 
    self.latent_dim = latent_dim 
    self.use_gpu = use_gpu

    self.number_of_ids = self.discriminator.number_of_ids

    # binary cross-entropy loss
    self.criterion_gan = nn.BCELoss()
    self.criterion_pose = nn.CrossEntropyLoss() # index is expected as target (and not one-hot)
    self.criterion_id = nn.CrossEntropyLoss()

    # move stuff to GPU if needed
    if self.use_gpu:
      self.discriminator.cuda()
      self.encoder.cuda()
      self.decoder.cuda()
      self.criterion_gan.cuda()
      self.criterion_pose.cuda()
      self.criterion_id.cuda()

  def get_fixed_data(self, dataset, plot=False):
    """
    Function to get fixed data, used to generate samples at every epoch.

    **Parameters**

    dataset: pyTorch Dataset
      The dataset where the image should come from.

    **Returns**

    fixed_image: pyTorch Variable
      The input image to encode/decode

    fixed_noise: pyTorch Variable
      The noise to inject in the decoder

    """
    pose = 0
    counter = 0
    while pose != 6:
      pose = dataset[counter]['pose']
      fixed_index = counter
      counter += 1
    fixed_image = torch.Tensor(torch.Size(self.image_size))
    fixed_image.copy_(dataset[counter]['image'])
  
    # plot the image if asked for
    if plot:
      fixed_id = dataset[counter]['id']
      fixed_pose = dataset[counter]['pose']
      from matplotlib import pyplot
      pyplot.title("ID -> {}, pose {}".format(fixed_id, fixed_pose))
      pyplot.imshow(numpy.rollaxis(numpy.rollaxis(((fixed_image.numpy()+1)/2.), 2),2))
      pyplot.show()
   
    # expand the fixed image, so that to have a batch for all possible poses
    fixed_image = fixed_image.expand(self.conditional_dim, self.image_size[0], self.image_size[1], self.image_size[2])
    fixed_image = Variable(fixed_image)
   
    # the noise to sample from
    fixed_noise = torch.FloatTensor(self.conditional_dim, self.noise_dim, 1, 1).normal_(0, 1)
    fixed_noise = Variable(fixed_noise)

    # build the set of one-hot encoded pose to sample from 
    fixed_one_hot = torch.FloatTensor(self.conditional_dim, self.conditional_dim, 1, 1).zero_()
    for k in range(self.conditional_dim):
      fixed_one_hot[k, k] = 1
    fixed_one_hot = Variable(fixed_one_hot)

    return fixed_image, fixed_noise, fixed_one_hot
   

  def load_models(self, models_dir):
    """
    Function to load existing models (last saved ones)
    
    **Parameters**

    model_dir: path
      dir where existing models are stored.
    
    **Returns**
    
    start_epoch: int
    """
    import os

    def getKey(item):
      """
      Return the time of last modification. Used to sort saved models

      **Parameters**

      item: file
        The file from which you woud get the modification time 
      
      **Returns**
        The last modification time of the file
      """
      return os.path.getmtime(item) 

    # to store model files
    discriminator_files = []
    encoder_files = []
    decoder_files = []
    start_epoch = 0
  
    # populate lists of models
    for f in os.listdir(models_dir):
      if 'discriminator' in f:
        discriminator_files.append(os.path.join(models_dir, f))
      if 'encoder' in f:
        encoder_files.append(os.path.join(models_dir, f))
      if 'decoder' in f:
        decoder_files.append(os.path.join(models_dir, f))

    # if some models already exists ...
    if len(discriminator_files) > 0:
    
      # sort according to modification time
      discriminator_files.sort(key=getKey)
      discriminator_file = discriminator_files[-1]
      encoder_file = encoder_files[-1]
      decoder_file = decoder_files[-1]

      # load models
      self.discriminator.load_state_dict(torch.load(discriminator_file, map_location=lambda storage, loc: storage))
      self.encoder.load_state_dict(torch.load(encoder_file, map_location=lambda storage, loc: storage))
      self.decoder.load_state_dict(torch.load(decoder_file, map_location=lambda storage, loc: storage))

      # get the start epoch
      start_epoch = int((discriminator_file.split("_")[-1]).split(".")[0]) + 1
      logger.info("Models from epoch {} loaded ! Starting at epoch {}".format(start_epoch - 1, start_epoch))
    else:
      logger.info("No exisiting models were found in {}".format(models_dir))
    return start_epoch


  def train(self, dataloader, n_epochs=10, learning_rate=0.0002, beta1=0.5, output_dir='out', plot=False, save_sample=10, pose_random=True, keep_model=1):
    """
    Function that performs the training.

    **Parameters**

    dataloader: pytorch DataLoader
      The dataloader for your data

    n_epochs: int
      The number of epochs you would like to train for

    learning_rate: float
      The learning rate for Adam optimizer

    beta1: float
      The beta1 for Adam optimizer

    output_dir: path
      The directory where you would like to output images and models

    plot: boolean
      If you want to plot some images during the training process (debug) 

    save_sample: int
      To save sample every X iterations

    pose_random: boolean
      To assign random pose labels to generated images (necessary actually)

    keep_model: int
      To keep model every X epochs (and the last one)
    """
    if not pose_random:
      logger.warn("Generating same poses as in training examples")

    # create directories
    images_dir = output_dir + "/images"
    models_dir = output_dir + "/models"
    log_dir = output_dir + "/logs"
    bob.io.base.create_directories_safe(images_dir)
    bob.io.base.create_directories_safe(models_dir)
    bob.io.base.create_directories_safe(log_dir)

    # check if some models already exists, in which case load the last ones 
    import os
    start_epoch = 0
    if os.path.isdir(models_dir):
      start_epoch = self.load_models(models_dir)

    # be sure to save samples at each epoch (at least)
    if save_sample >= len(dataloader):
      save_sample = len(dataloader) - 1

    # labels for real/fake
    real_label = 1
    fake_label = 0
    
    # setup optimizer
    generator_params = list(self.encoder.parameters()) + list(self.decoder.parameters())
    optimizerD = optim.Adam(self.discriminator.parameters(), lr=learning_rate, betas=(beta1, 0.999))
    optimizerG = optim.Adam(generator_params, lr=learning_rate, betas=(beta1, 0.999))

    # get fixed data to sample from after each epoch
    fixed_image, fixed_noise, fixed_one_hot = self.get_fixed_data(dataloader.dataset, plot)

    # statistics - losses
    discriminator_loss = []
    generator_loss = []

    # ================
    # === LET'S GO ===
    # ================
    for epoch in range(start_epoch, n_epochs):
      for i, data in enumerate(dataloader, 0):
    
        
        start = time.time()

        # get the batch data, pose and id labels
        real_images = data['image']
        poses = data['pose']
        ids = data['id']

        # sanity check
        if max(ids) >= self.number_of_ids:
          logger.error("Something is wrong here: I have an ID with index {}, and the number of IDs is {}".format(max(ids), self.number_of_ids))
          import sys
          sys.exit()

        # WARNING: the last batch could be smaller than the provided size 
        # (you could avoid that by setting the drop_last flag to True in dataloader constructor)
        batch_size = len(real_images)
        
        # generate random pose index - to train to generate "unknown" poses !
        random_poses = numpy.random.randint(0, self.conditional_dim, batch_size)
        random_poses = torch.LongTensor(random_poses)
       
        # get a minibatch of noise 
        noise = torch.FloatTensor(batch_size, self.noise_dim, 1, 1).normal_(0, 1)

        # create the one hot conditional vector on pose (decoder) 
        one_hot_vector = torch.FloatTensor(batch_size, self.conditional_dim, 1, 1).zero_()
        for k in range(batch_size):
          if pose_random:
            one_hot_vector[k, int(random_poses[k])] = 1
          else:
            one_hot_vector[k, poses[k]] = 1
        
        # label for fake/real
        label_gan = torch.FloatTensor(batch_size)
        
        # move stuff to GPU if needed
        if self.use_gpu:
          # inputs
          real_images = real_images.cuda()
          noise = noise.cuda()
          one_hot_vector = one_hot_vector.cuda()
          #labels
          label_gan = label_gan.cuda()
          poses = poses.cuda()
          random_poses = random_poses.cuda()
          ids = ids.cuda()

        # =============
        # DISCRIMINATOR 
        # =============
        self.discriminator.zero_grad()
     
        # === REAL DATA ===
        imagev = Variable(real_images)
        label_gan.resize_(batch_size).fill_(real_label)
        label_gan_v = Variable(label_gan)
        label_pose_v = Variable(poses)
        label_id_v = Variable(ids)
        output_real = self.discriminator(imagev)

        errD_real_id = self.criterion_id(output_real[:, :self.number_of_ids], label_id_v)
        errD_real_pose = self.criterion_pose(output_real[:, self.number_of_ids:(self.number_of_ids + self.conditional_dim)], label_pose_v)
        errD_real_gan = self.criterion_gan(output_real[:, -1], label_gan_v)

        errD_real_id.backward(retain_graph=True)
        errD_real_pose.backward(retain_graph=True)
        errD_real_gan.backward(retain_graph=True)

        # === FAKE DATA ===
        noisev = Variable(noise)
        one_hot_vv = Variable(one_hot_vector)
        encoded_ids = self.encoder(imagev)
        fake = self.decoder(noisev, one_hot_vv, encoded_ids)
        label_gan_v = Variable(label_gan.fill_(fake_label))
        label_random_poses_v = Variable(random_poses)
        output_fake = self.discriminator(fake)
        
        errD_fake_id = self.criterion_id(output_fake[:, :self.number_of_ids], label_id_v)
        if pose_random:
          errD_fake_pose = self.criterion_pose(output_fake[:, self.number_of_ids:(self.number_of_ids + self.conditional_dim)], label_random_poses_v)
        else:
          errD_fake_pose = self.criterion_pose(output_fake[:, self.number_of_ids:(self.number_of_ids + self.conditional_dim)], label_pose_v)
        errD_fake_gan = self.criterion_gan(output_fake[:, -1], label_gan_v)

        errD_fake_id.backward(retain_graph=True)
        errD_fake_pose.backward(retain_graph=True)
        errD_fake_gan.backward(retain_graph=True)
        
        # perform optimization (i.e. update discriminator parameters)
        errD = errD_real_id + errD_real_pose + (errD_real_gan + errD_fake_gan)
        optimizerD.step()
       
        # =========
        # GENERATOR 
        # =========
        self.encoder.zero_grad()
        self.decoder.zero_grad()
        label_gan_v = Variable(label_gan.fill_(real_label))  # fake labels are real for generator cost
        output_generated = self.discriminator(fake)
        
        errG_id = self.criterion_id(output_generated[:, :self.number_of_ids], label_id_v)
        if pose_random:
          errG_pose = self.criterion_pose(output_generated[:, self.number_of_ids:(self.number_of_ids + self.conditional_dim)], label_random_poses_v)
        else:
          errG_pose = self.criterion_pose(output_generated[:, self.number_of_ids:(self.number_of_ids + self.conditional_dim)], label_pose_v)
        errG_gan = self.criterion_gan(output_generated[:, -1], label_gan_v)
        
        errG_id.backward(retain_graph=True)
        errG_pose.backward(retain_graph=True)
        errG_gan.backward(retain_graph=True)
        
        # perform optimization (i.e. update discriminator parameters)
        errG = errG_id + errG_pose + errG_gan 
        optimizerG.step()
        # === END OF OPTIMIZATION ===
        
        # log stuff
        end = time.time()
        logger.info("[{}/{}][{}/{}] => Loss D = {} -- Loss G = {} (time spent: {})".format(epoch, n_epochs, i, len(dataloader), errD.data[0], errG.data[0], (end-start)))
        discriminator_loss.append(errD.data[0])
        generator_loss.append(errG.data[0])
        self.check_batch_statistics(output_real, output_fake, output_generated, ids, poses, random_poses, batch_size, log_dir)

        # ==========
        # SAVE STUFF 
        # ==========
        if ((i % save_sample) == 0) and (i > 0):
          
          # get a random example in this minibatch
          index = numpy.random.randint(0, batch_size)
          logger.info("Saving example {} in this batch (epoch {} - iteration {})".format(index, epoch, i))
          
          # move stuff back to CPU (needed to use numpy())
          real_images = real_images.cpu()
          fake = fake.cpu()
          real_example = (real_images[index].numpy() + 1)/2.
          generated_example = (fake[index].data.numpy() + 1)/2.
          id_example = ids[index]
          pose_example = poses[index]
          if pose_random:
            pose_example = random_poses[index]

          # save hdf5 data:
          filename = images_dir + '/generated_data_{}_{}.hdf5'.format(epoch, i)
          f = bob.io.base.HDF5File(filename, 'w')
          f.set('id', id_example)
          f.set('real_example', real_example)
          f.set('generated_example', generated_example)
          f.set('real_pose', poses[index])
          f.set('target_pose', pose_example)
          del f

          # save losses
          filename = log_dir + '/losses_{}_{}.hdf5'.format(epoch, i)
          f = bob.io.base.HDF5File(filename, 'a')
          f.set('d_loss', discriminator_loss)
          f.set('g_loss', generator_loss)
          del f

          # remove the previous losses file
          import glob, os
          losses_files = glob.glob(log_dir + '/losses_*') 
          for loss_file in losses_files:
            if loss_file != filename:
              os.remove(loss_file)

      # save generated images at every epoch
      # TODO: model moved to CPU and back and I don't really know why (expected CPU tensor error)
      # To summarize:
      # tried to move tensors, variables on the GPU -> does not work
      # let the tensors on the CPU -> does not work
      # => model has to be brought back to the CPU :/
      self.encoder = self.encoder.cpu()
      self.decoder = self.decoder.cpu()
      
      fixed_encoded_id = self.encoder(fixed_image)
      fake_examples = self.decoder(fixed_noise, fixed_one_hot, fixed_encoded_id)
      to_save = torch.Tensor(torch.Size((self.conditional_dim + 1, self.image_size[0], self.image_size[1], self.image_size[2])))
      to_save[0] = fixed_image.data[0]
      to_save[1:] = fake_examples.data
      vutils.save_image(to_save, '%s/fake_samples_epoch_%03d.png' % (images_dir, epoch), normalize=True)
      
      if torch.cuda.is_available():
        self.encoder = self.encoder.cuda()
        self.decoder = self.decoder.cuda()

      # do checkpointing
      torch.save(self.encoder.state_dict(), '%s/encoder_epoch_%d.pth' % (models_dir, epoch))
      torch.save(self.decoder.state_dict(), '%s/decoder_epoch_%d.pth' % (models_dir, epoch))
      torch.save(self.discriminator.state_dict(), '%s/discriminator_epoch_%d.pth' % (models_dir, epoch))
      
      # remove models that we don't want to keep
      import glob, os
      model_files = glob.glob(models_dir + '/*.pth') 
      for model_file in model_files:
        model_epoch = int(model_file.split('_')[-1].split('.')[0])
        if ((model_epoch % keep_model) != 0) and (model_epoch != epoch):
          os.remove(model_file)
          logger.info("{} removed !".format(model_file))


  def check_batch_statistics(self, output_real, output_fake, output_generator, 
                             label_id, label_pose, label_random_pose, batch_size, log_dir):
    """
    Compute some performance stats on the current mini-batch

    **Parameters**

    output_real: pyTorch Tensor (batch_size, (#of ids + dim(condition) + 1))
      The discriminator output for real images

    output_fake: pyTorch Tensor (batch_size, (#of ids + dim(condition) + 1))
      The discriminator output for fake examples

    output_generator: pyTorch Tensor (batch_size, (#of ids + dim(condition) + 1))
      The discriminator output used to train the generator

    label_id: pyTorch Tensor (batch_size)
      The label for the identity 

    label_pose: pyTorch Tensor (batch_size)
      The (real) label for the pose 

    label_random_pose: pyTorch Tensor (batch_size)
      The random label for the pose 

    batch_size: int
      The size of the current batch 

    log_dir: path
      The directory where to store the logs 
    """

    # --- REAL ---
    output_real_id = output_real[:, :self.number_of_ids]
    output_real_pose = output_real[:, self.number_of_ids:(self.number_of_ids + self.conditional_dim)]
    output_real_gan = output_real[:, -1]

    # --- FAKE ---
    output_fake_id = output_fake[:, :self.number_of_ids]
    output_fake_pose = output_fake[:, self.number_of_ids:(self.number_of_ids + self.conditional_dim)]
    output_fake_gan = output_fake[:, -1]
    
    # --- GENERATOR ---
    output_generator_id = output_generator[:, :self.number_of_ids]
    output_generator_pose = output_generator[:, self.number_of_ids:(self.number_of_ids + self.conditional_dim)]
    output_generator_gan = output_generator[:, -1]

    real_id_correct = 0
    real_pose_correct = 0
    real_gan_correct = 0
    
    fake_id_correct = 0
    fake_pose_correct = 0
    fake_gan_correct = 0
    
    generator_id_correct = 0
    generator_pose_correct = 0
    generator_gan_correct = 0
    
    for i in range(batch_size): 

      prob, inferred_real_id = torch.max(output_real_id[i], 0)
      prob, inferred_real_pose = torch.max(output_real_pose[i], 0)
      prob_real = output_real_gan[i].data
        
      if inferred_real_id.data[0] == label_id[i]:
        real_id_correct += 1
      if inferred_real_pose.data[0] == label_pose[i]:
        real_pose_correct += 1
      if prob_real[0] > 0.5:
        real_gan_correct +=1


      prob, inferred_fake_id = torch.max(output_fake_id[i], 0)
      prob, inferred_fake_pose = torch.max(output_fake_pose[i], 0)
      prob_real = output_fake_gan[i].data
        
      if inferred_fake_id.data[0] == label_id[i]:
        fake_id_correct += 1
      if inferred_fake_pose.data[0] == label_pose[i]:
        fake_pose_correct += 1
      if prob_real[0] < 0.5:
        fake_gan_correct +=1

      
      prob, inferred_generator_id = torch.max(output_generator_id[i], 0)
      prob, inferred_generator_pose = torch.max(output_generator_pose[i], 0)
      prob_real = output_generator_gan[i].data
        
      if inferred_generator_id.data[0] == label_id[i]:
        generator_id_correct += 1
      if inferred_generator_pose.data[0] == label_pose[i]:
        generator_pose_correct += 1
      if prob_real[0] < 0.5:
        generator_gan_correct +=1
   
    fd_real = bob.io.base.HDF5File(log_dir + '/discriminator_real_stats.hdf5', 'a')
    fd_fake = bob.io.base.HDF5File(log_dir + '/discriminator_fake_stats.hdf5', 'a')
    fg = bob.io.base.HDF5File(log_dir + '/generator_stats.hdf5', 'a')
    
    fd_real.append('r_id_accuracy', real_id_correct/float(batch_size))
    fd_real.append('r_pose_accuracy', real_pose_correct/float(batch_size))
    fd_real.append('r_real_accuracy', real_gan_correct/float(batch_size))

    fd_fake.append('f_id_accuracy', fake_id_correct/float(batch_size))
    fd_fake.append('f_pose_accuracy', fake_pose_correct/float(batch_size))
    fd_fake.append('f_fake_accuracy', fake_gan_correct/float(batch_size))

    fg.append('g_id_accuracy', fake_id_correct/float(batch_size))
    fg.append('g_pose_accuracy', fake_pose_correct/float(batch_size))
    fg.append('g_fake_accuracy', real_gan_correct/float(batch_size))

    del fd_real
    del fd_fake
    del fg

    logger.debug("[REAL] accuracy on ID = {} ({}/{})".format(real_id_correct/float(batch_size), real_id_correct, batch_size))
    logger.debug("[REAL] accuracy on pose = {} ({}/{})".format(real_pose_correct/float(batch_size), real_pose_correct, batch_size))
    logger.debug("[REAL] accuracy on real/fake = {} ({}/{})".format(real_gan_correct/float(batch_size), real_gan_correct, batch_size))
    
    logger.debug("[FAKE] accuracy on ID = {} ({}/{})".format(fake_id_correct/float(batch_size), fake_id_correct, batch_size))
    logger.debug("[FAKE] accuracy on pose = {} ({}/{})".format(fake_pose_correct/float(batch_size), fake_pose_correct, batch_size))
    logger.debug("[FAKE] accuracy on real/fake = {} ({}/{})".format(fake_gan_correct/float(batch_size), fake_gan_correct, batch_size))

    logger.debug("[GENERATOR] accuracy on ID = {} ({}/{})".format(generator_id_correct/float(batch_size), generator_id_correct, batch_size))
    logger.debug("[GENERATOR] accuracy on pose = {} ({}/{})".format(generator_pose_correct/float(batch_size), generator_pose_correct, batch_size))
    logger.debug("[GENERATOR] accuracy on real/fake = {} ({}/{})".format(generator_gan_correct/float(batch_size), generator_gan_correct, batch_size))
